# Copyright 2016 Google Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
# in compliance with the License. You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License
# is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
# or implied. See the License for the specific language governing permissions and limitations under
# the License.

"""IPython Functionality for the Google Monitoring API."""
from __future__ import absolute_import

try:
  import IPython.core.display
  import IPython.core.magic
except ImportError:
  raise Exception('This module can only be loaded in ipython.')

import datalab.stackdriver.monitoring as gcm
import datalab.utils.commands


@IPython.core.magic.register_line_cell_magic
def monitoring(line, cell=None):
  """Implements the monitoring cell magic for ipython notebooks.

  Args:
    line: the contents of the storage line.
  Returns:
    The results of executing the cell.
  """
  parser = datalab.utils.commands.CommandParser(prog='monitoring', description=(
      'Execute various Monitoring-related operations. Use "%monitoring '
      '<command> -h" for help on a specific command.'))

  list_parser = parser.subcommand(
      'list', 'List the metrics or resource types in a monitored project.')

  list_metric_parser = list_parser.subcommand(
      'metrics',
      'List the metrics that are available through the Monitoring API.')
  list_metric_parser.add_argument(
      '-t', '--type',
      help='The type of metric(s) to list; can include wildchars.')
  list_metric_parser.add_argument(
      '-p', '--project', help='The project on which to execute the request.')
  list_metric_parser.set_defaults(func=_list_metric_descriptors)

  list_resource_parser = list_parser.subcommand(
      'resource_types',
      ('List the monitored resource types that are available through the '
       'Monitoring API.'))
  list_resource_parser.add_argument(
      '-p', '--project', help='The project on which to execute the request.')
  list_resource_parser.add_argument(
      '-t', '--type',
      help='The resource type(s) to list; can include wildchars.')
  list_resource_parser.set_defaults(func=_list_resource_descriptors)

  list_group_parser = list_parser.subcommand(
      'groups',
      ('List the Stackdriver groups in this project.'))
  list_group_parser.add_argument(
      '-p', '--project', help='The project on which to execute the request.')
  list_group_parser.add_argument(
      '-n', '--name',
      help='The name of the group(s) to list; can include wildchars.')
  list_group_parser.set_defaults(func=_list_groups)

  return datalab.utils.commands.handle_magic_line(line, cell, parser)


def _list_metric_descriptors(args, _):
  """Lists the metric descriptors in the project."""
  project_id = args['project']
  pattern = args['type'] or '*'
  descriptors = gcm.MetricDescriptors(project_id=project_id)
  dataframe = descriptors.as_dataframe(pattern=pattern)
  return _render_dataframe(dataframe)


def _list_resource_descriptors(args, _):
  """Lists the resource descriptors in the project."""
  project_id = args['project']
  pattern = args['type'] or '*'
  descriptors = gcm.ResourceDescriptors(project_id=project_id)
  dataframe = descriptors.as_dataframe(pattern=pattern)
  return _render_dataframe(dataframe)


def _list_groups(args, _):
  """Lists the groups in the project."""
  project_id = args['project']
  pattern = args['name'] or '*'
  groups = gcm.Groups(project_id=project_id)
  dataframe = groups.as_dataframe(pattern=pattern)
  return _render_dataframe(dataframe)


def _render_dataframe(dataframe):
  """Helper to render a dataframe as an HTML table."""
  data = dataframe.to_dict(orient='records')
  fields = dataframe.columns.tolist()
  return IPython.core.display.HTML(
      datalab.utils.commands.HtmlBuilder.render_table(data, fields))



# Copyright 2015 Google Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
# in compliance with the License. You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License
# is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
# or implied. See the License for the specific language governing permissions and limitations under
# the License.

"""Implements Table and View Schema APIs."""
from __future__ import absolute_import
from __future__ import unicode_literals
from builtins import str
from builtins import range
from past.builtins import basestring
from builtins import object

import datetime
import pandas


class Schema(list):
  """Represents the schema of a BigQuery table as a flattened list of objects representing fields.

   Each field object has name, data_type, mode and description properties. Nested fields
   get flattened with their full-qualified names. So a Schema that has an object A with nested
   field B will be represented as [(name: 'A', ...), (name: 'A.b', ...)].
  """

  class Field(object):
    """ Represents a single field in a Table schema.

    This has the properties:

    - name: the flattened, full-qualified name of the field.
    - data_type: the type of the field as a string ('INTEGER', 'BOOLEAN', 'FLOAT', 'STRING'
       or 'TIMESTAMP').
    - mode: the mode of the field; 'NULLABLE' by default.
    - description: a description of the field, if known; empty string by default.

     """

    # TODO(gram): consider renaming data_type member to type. Yes, it shadows top-level
    # name but that is what we are using in __str__ and __getitem__ and is what is used in BQ.
    # The shadowing is unlikely to cause problems.
    def __init__(self, name, data_type, mode='NULLABLE', description=''):
      self.name = name
      self.data_type = data_type
      self.mode = mode
      self.description = description

    def _repr_sql_(self):
      """Returns a representation of the field for embedding into a SQL statement.

      Returns:
        A formatted field name for use within SQL statements.
      """
      return self.name

    def __eq__(self, other):
      """ Compare two schema field objects for equality (ignoring description). """
      return self.name == other.name and self.data_type == other.data_type\
          and self.mode == other.mode

    def __str__(self):
      """ Returns the schema field as a string form of a dictionary. """
      return "{ 'name': '%s', 'type': '%s', 'mode':'%s', 'description': '%s' }" % \
             (self.name, self.data_type, self.mode, self.description)

    def __repr__(self):
      """ Returns the schema field as a string form of a dictionary. """
      return str(self)

    def __getitem__(self, item):
      # TODO(gram): Currently we need this for a Schema object to work with the Parser object.
      # Eventually if we change Parser to only work with Schema (and not also with the
      # schema dictionaries in query results) we can remove this.

      if item == 'name':
        return self.name
      if item == 'type':
        return self.data_type
      if item == 'mode':
        return self.mode
      if item == 'description':
        return self.description

  @staticmethod
  def _from_dataframe(dataframe, default_type='STRING'):
    """
      Infer a BigQuery table schema from a Pandas dataframe. Note that if you don't explicitly set
      the types of the columns in the dataframe, they may be of a type that forces coercion to
      STRING, so even though the fields in the dataframe themselves may be numeric, the type in the
      derived schema may not be. Hence it is prudent to make sure the Pandas dataframe is typed
      correctly.

    Args:
      dataframe: The DataFrame.
      default_type : The default big query type in case the type of the column does not exist in
          the schema. Defaults to 'STRING'.
    Returns:
      A list of dictionaries containing field 'name' and 'type' entries, suitable for use in a
          BigQuery Tables resource schema.
    """

    type_mapping = {
      'i': 'INTEGER',
      'b': 'BOOLEAN',
      'f': 'FLOAT',
      'O': 'STRING',
      'S': 'STRING',
      'U': 'STRING',
      'M': 'TIMESTAMP'
    }

    fields = []
    for column_name, dtype in dataframe.dtypes.iteritems():
      fields.append({'name': column_name,
                     'type': type_mapping.get(dtype.kind, default_type)})

    return fields

  @staticmethod
  def from_dataframe(dataframe, default_type='STRING'):
    """
      Infer a BigQuery table schema from a Pandas dataframe. Note that if you don't explicitly set
      the types of the columns in the dataframe, they may be of a type that forces coercion to
      STRING, so even though the fields in the dataframe themselves may be numeric, the type in the
      derived schema may not be. Hence it is prudent to make sure the Pandas dataframe is typed
      correctly.

    Args:
      dataframe: The DataFrame.
      default_type : The default big query type in case the type of the column does not exist in
          the schema. Defaults to 'STRING'.
    Returns:
      A Schema.
    """
    return Schema(Schema._from_dataframe(dataframe, default_type=default_type))

  @staticmethod
  def _get_field_entry(name, value):
    entry = {'name': name}
    if isinstance(value, datetime.datetime):
      _type = 'TIMESTAMP'
    elif isinstance(value, bool):
      _type = 'BOOLEAN'
    elif isinstance(value, float):
      _type = 'FLOAT'
    elif isinstance(value, int):
      _type = 'INTEGER'
    elif isinstance(value, dict) or isinstance(value, list):
      _type = 'RECORD'
      entry['fields'] = Schema._from_record(value)
    else:
      _type = 'STRING'
    entry['type'] = _type
    return entry

  @staticmethod
  def _from_dict_record(data):
    """
    Infer a BigQuery table schema from a dictionary. If the dictionary has entries that
    are in turn OrderedDicts these will be turned into RECORD types. Ideally this will
    be an OrderedDict but it is not required.

    Args:
      data: The dict to infer a schema from.
    Returns:
      A list of dictionaries containing field 'name' and 'type' entries, suitable for use in a
          BigQuery Tables resource schema.
    """
    return [Schema._get_field_entry(name, value) for name, value in list(data.items())]

  @staticmethod
  def _from_list_record(data):
    """
    Infer a BigQuery table schema from a list of values.

    Args:
      data: The list of values.
    Returns:
      A list of dictionaries containing field 'name' and 'type' entries, suitable for use in a
          BigQuery Tables resource schema.
    """
    return [Schema._get_field_entry('Column%d' % (i + 1), value) for i, value in enumerate(data)]

  @staticmethod
  def _from_record(data):
    """
    Infer a BigQuery table schema from a list of fields or a dictionary. The typeof the elements
    is used. For a list, the field names are simply 'Column1', 'Column2', etc.

    Args:
      data: The list of fields or dictionary.
    Returns:
      A list of dictionaries containing field 'name' and 'type' entries, suitable for use in a
          BigQuery Tables resource schema.
    """
    if isinstance(data, dict):
      return Schema._from_dict_record(data)
    elif isinstance(data, list):
      return Schema._from_list_record(data)
    else:
      raise Exception('Cannot create a schema from record %s' % str(data))

  @staticmethod
  def from_record(source):
    """
    Infers a table/view schema from a single record that can contain a list of fields or a
    dictionary of fields. The type of the elements is used for the types in the schema. For a
    dict, key names are used for column names while for a list, the field names are simply named
    'Column1', 'Column2', etc. Note that if using a dict you may want to use an OrderedDict
    to ensure column ordering is deterministic.

    Args:
      source: The list of field values or dictionary of key/values.

    Returns:
      A Schema for the data.
    """
    # TODO(gram): may want to allow an optional second argument which is a list of field
    # names; could be useful for the record-containing-list case.
    return Schema(Schema._from_record(source))

  @staticmethod
  def from_data(source):
    """Infers a table/view schema from its JSON representation, a list of records, or a Pandas
       dataframe.

    Args:
      source: the Pandas Dataframe, a dictionary representing a record, a list of heterogeneous
          data (record) or homogeneous data (list of records) from which to infer the schema, or
          a definition of the schema as a list of dictionaries with 'name' and 'type' entries
          and possibly 'mode' and 'description' entries. Only used if no data argument was provided.
          'mode' can be 'NULLABLE', 'REQUIRED' or 'REPEATED'. For the allowed types, see:
          https://cloud.google.com/bigquery/preparing-data-for-bigquery#datatypes

          Note that there is potential ambiguity when passing a list of lists or a list of
          dicts between whether that should be treated as a list of records or a single record
          that is a list. The heuristic used is to check the length of the entries in the
          list; if they are equal then a list of records is assumed. To avoid this ambiguity
          you can instead use the Schema.from_record method which assumes a single record,
          in either list of values or dictionary of key-values form.

    Returns:
      A Schema for the data.
    """
    if isinstance(source, pandas.DataFrame):
      bq_schema = Schema._from_dataframe(source)
    elif isinstance(source, list):
      if len(source) == 0:
        bq_schema = source
      elif all(isinstance(d, dict) for d in source):
        if all('name' in d and 'type' in d for d in source):
          # It looks like a bq_schema; use it as-is.
          bq_schema = source
        elif all(len(d) == len(source[0]) for d in source):
          bq_schema = Schema._from_dict_record(source[0])
        else:
          raise Exception(('Cannot create a schema from heterogeneous list %s; perhaps you meant ' +
                          'to use Schema.from_record?') % str(source))
      elif isinstance(source[0], list) and \
          all([isinstance(l, list) and len(l) == len(source[0]) for l in source]):
        # A list of lists all of the same length; treat first entry as a list record.
        bq_schema = Schema._from_record(source[0])
      else:
        # A heterogeneous list; treat as a record.
        raise Exception(('Cannot create a schema from heterogeneous list %s; perhaps you meant ' +
                        'to use Schema.from_record?') % str(source))
    elif isinstance(source, dict):
      raise Exception(('Cannot create a schema from dict %s; perhaps you meant to use ' +
                      'Schema.from_record?') % str(source))
    else:
      raise Exception('Cannot create a schema from %s' % str(source))
    return Schema(bq_schema)

  def __init__(self, definition=None):
    """Initializes a Schema from its raw JSON representation, a Pandas Dataframe, or a list.

    Args:
      definition: a definition of the schema as a list of dictionaries with 'name' and 'type'
          entries and possibly 'mode' and 'description' entries. Only used if no data argument was
          provided. 'mode' can be 'NULLABLE', 'REQUIRED' or 'REPEATED'. For the allowed types, see:
          https://cloud.google.com/bigquery/preparing-data-for-bigquery#datatypes
    """
    super(Schema, self).__init__()
    self._map = {}
    self._bq_schema = definition
    self._populate_fields(definition)

  def __getitem__(self, key):
    """Provides ability to lookup a schema field by position or by name.
    """
    if isinstance(key, basestring):
      return self._map.get(key, None)
    # noinspection PyCallByClass
    return list.__getitem__(self, key)

  def _add_field(self, name, data_type, mode='NULLABLE', description=''):
    field = Schema.Field(name, data_type, mode, description)
    self.append(field)
    self._map[name] = field

  def find(self, name):
    """ Get the index of a field in the flattened list given its (fully-qualified) name.

    Args:
      name: the fully-qualified name of the field.
    Returns:
      The index of the field, if found; else -1.
    """
    for i in range(0, len(self)):
      if self[i].name == name:
        return i
    return -1

  def _populate_fields(self, data, prefix=''):
    for field_data in data:
      name = prefix + field_data['name']
      data_type = field_data['type']
      self._add_field(name, data_type, field_data.get('mode', None),
                      field_data.get('description', None))

      if data_type == 'RECORD':
        # Recurse into the nested fields, using this field's name as a prefix.
        self._populate_fields(field_data.get('fields'), name + '.')

  def __str__(self):
    """ Returns a string representation of the non-flattened form of the schema. """
    # TODO(gram): We should probably return the flattened form. There was a reason why this is
    # not but I don't remember what it was. Figure that out and fix this.
    return str(self._bq_schema)

  def __eq__(self, other):
    """ Compares two schema for equality. """
    other_map = other._map
    if len(self._map) != len(other_map):
      return False
    for name in self._map.keys():
      if name not in other_map:
        return False
      if not self._map[name] == other_map[name]:
        return False
    return True

  def __ne__(self, other):
    """ Compares two schema for inequality. """
    return not(self.__eq__(other))



# Copyright 2016 Google Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
# in compliance with the License. You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License
# is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
# or implied. See the License for the specific language governing permissions and limitations under
# the License.


import collections
import json
from numbers import Number
import numpy
import os
import pandas as pd

import google.cloud.ml as ml

from . import _metadata


class LocalPredictor(object):
  """Preforms local predictions on given data.
  """

  def __init__(self, model_dir, label_output=None):
    """Initializes an instance of LocalPredictor.

    Args:
      model_dir: a directory that contains model checkpoint and metagraph. Can be local or GCS.
      label_output: the name of the output column where all values should be converted from
          index to labels. Only useful in classification. If specified, a metadata.yaml file is required.
    """
    self._model_dir = model_dir
    self._metadata_path = None
    self._metadata = None
    metadata_path = os.path.join(model_dir, "metadata.yaml")
    if ml.util._file.file_exists(metadata_path):
      self._metadata_path = metadata_path
      self._metadata = _metadata.Metadata(metadata_path)
    self._label_output = label_output

  def predict(self, data):
    """Make predictions on given data.

    Args:
      data: a list of feature data or a pandas DataFrame. Each element in the list is an instance
          which is a dictionary of feature data.
          An example:
            [{"sepal_length": 4.9, "sepal_width": 2.5, "petal_length": 4.5, "petal_width": 1.7},
             {"sepal_length": 5.7, "sepal_width": 2.8, "petal_length": 4.1, "petal_width": 1.3}]

    Returns:
      A list of prediction results for given instances. Each element is a dictionary representing
          output mapping from the graph.
      An example:
        [{"predictions": 1, "score": [0.00078, 0.71406, 0.28515]},
         {"predictions": 1, "score": [0.00244, 0.99634, 0.00121]}]

    Raises: Exception if the prediction result has incorrect label types
    """
    session, _ = ml.session_bundle.load_session_bundle_from_path(self._model_dir)
    # get the mappings between aliases and tensor names for both inputs and outputs
    input_key = json.loads(session.graph.get_collection(
        ml.session_bundle.INPUTS_KEY)[0]).values()[0]
    output_alias_map = json.loads(session.graph.get_collection(ml.session_bundle.OUTPUTS_KEY)[0])
    aliases, tensornames = zip(*output_alias_map.items())

    if isinstance(data, pd.DataFrame):
      data = data.T.to_dict().values()

    feed_dict = collections.defaultdict(list)
    if self._metadata_path is not None:
      transformer = ml.features.FeatureProducer(self._metadata_path)
      for instance in data:
        preprocessed = transformer.preprocess(instance)
        feed_dict[input_key].append(preprocessed.SerializeToString())
    else:
      for instance in data:
        feed_dict[input_key].append(json.dumps(instance))

    result = session.run(fetches=tensornames, feed_dict=feed_dict)
    if len(result) == 1:
      result = [result]
    predictions = []
    for row in zip(*result):
      prediction_row = {}
      for name, value in zip(aliases, row):
        if (self._metadata is not None and self._label_output is not None
            and name == self._label_output):
          if not isinstance(value, Number):
            raise Exception('Cannot get labels because output "%s" is type %s but not number.'
                % (name, type(value)))
          prediction_row[name] = str(self._metadata.get_classification_label(value)) + \
              (' (%d)' % value)
        elif isinstance(value, numpy.generic):
          prediction_row[name] = numpy.asscalar(value)
        elif isinstance(value, numpy.ndarray):
          prediction_row[name] = value.tolist()
        else:
          prediction_row[name] = value
      predictions.append(prediction_row)
    return predictions



# Copyright 2015 Google Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
# in compliance with the License. You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License
# is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
# or implied. See the License for the specific language governing permissions and limitations under
# the License.

"""Google Cloud Platform library - Generic SQL Helpers."""
from __future__ import absolute_import
from __future__ import unicode_literals
import types


# Names used for the arg parser, unnamed (main) query and last query in the module.
# Note that every module has a last query, but not every module has a main query.

_SQL_MODULE_ARGPARSE = '_sql_module_arg_parser'
_SQL_MODULE_MAIN = '_sql_module_main'
_SQL_MODULE_LAST = '_sql_module_last'


def get_default_query_from_module(module):
  """ Given a %%sql module return the default (last) query for the module.

  Args:
    module: the %%sql module.

  Returns:
    The default query associated with this module.
  """
  if isinstance(module, types.ModuleType):
    return module.__dict__.get(_SQL_MODULE_LAST, None)
  return None


def _next_token(sql):
  """ This is a basic tokenizer for our limited purposes.
  It splits a SQL statement up into a series of segments, where a segment is one of:
  - identifiers
  - left or right parentheses
  - multi-line comments
  - single line comments
  - white-space sequences
  - string literals
  - consecutive strings of characters that are not one of the items above

  The aim is for us to be able to find function calls (identifiers followed by '('), and the
  associated closing ')') so we can augment these if needed.

  Args:
    sql: a SQL statement as a (possibly multi-line) string.

  Returns:
    For each call, the next token in the initial input.
  """
  i = 0

  # We use some lambda's to make the logic more clear. The start_* functions return
  # true if i is the index of the start of that construct, while the end_* functions
  # return true if i point to the first character beyond that construct or the end of the
  # content.
  #
  # We don't currently need numbers so the tokenizer here just does sequences of
  # digits as a convenience to shrink the total number of tokens. If we needed numbers
  # later we would need a special handler for these much like strings.

  start_multi_line_comment = lambda s, i: s[i] == '/' and i < len(s) - 1 and s[i + 1] == '*'
  end_multi_line_comment = lambda s, i: s[i - 2] == '*' and s[i - 1] == '/'
  start_single_line_comment = lambda s, i: s[i] == '-' and i < len(s) - 1 and s[i + 1] == '-'
  end_single_line_comment = lambda s, i: s[i - 1] == '\n'
  start_whitespace = lambda s, i: s[i].isspace()
  end_whitespace = lambda s, i: not s[i].isspace()
  start_number = lambda s, i: s[i].isdigit()
  end_number = lambda s, i: not s[i].isdigit()
  start_identifier = lambda s, i: s[i].isalpha() or s[i] == '_' or s[i] == '$'
  end_identifier = lambda s, i: not(s[i].isalnum() or s[i] == '_')
  start_string = lambda s, i: s[i] == '"' or s[i] == "'"
  always_true = lambda s, i: True

  while i < len(sql):
    start = i
    if start_multi_line_comment(sql, i):
      i += 1
      end_checker = end_multi_line_comment
    elif start_single_line_comment(sql, i):
      i += 1
      end_checker = end_single_line_comment
    elif start_whitespace(sql, i):
      end_checker = end_whitespace
    elif start_identifier(sql, i):
      end_checker = end_identifier
    elif start_number(sql, i):
      end_checker = end_number
    elif start_string(sql, i):
      # Special handling here as we need to check for escaped closing quotes.
      quote = sql[i]
      end_checker = always_true
      i += 1
      while i < len(sql) and sql[i] != quote:
        i += 2 if sql[i] == '\\' else 1
    else:
      # We return single characters for everything else
      end_checker = always_true

    i += 1
    while i < len(sql) and not end_checker(sql, i):
      i += 1

    (yield sql[start:i])


def tokenize(sql):
  """ This is a basic tokenizer for our limited purposes.
  It splits a SQL statement up into a series of segments, where a segment is one of:
  - identifiers
  - left or right parentheses
  - multi-line comments
  - single line comments
  - white-space sequences
  - string literals
  - consecutive strings of characters that are not one of the items above

  The aim is for us to be able to find function calls (identifiers followed by '('), and the
  associated closing ')') so we can augment these if needed.

  Args:
    sql: a SQL statement as a (possibly multi-line) string.

  Returns:
    A list of strings corresponding to the groups above.
  """
  return list(_next_token(sql))




# Copyright 2016 Google Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
# in compliance with the License. You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License
# is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
# or implied. See the License for the specific language governing permissions and limitations under
# the License.


import argparse
import psutil
import subprocess
import time

import datalab.utils
import datalab.storage


class TensorBoardManager(object):
  """Start, shutdown, and list TensorBoard instances.
  """

  @staticmethod
  def get_running_list():
    """List running TensorBoard instances.

    Returns: A list of {'pid': pid, 'logdir': logdir, 'port': port}
    """
    running_list = []
    parser = argparse.ArgumentParser()
    parser.add_argument('--logdir')
    parser.add_argument('--port')
    for p in psutil.process_iter():
      if p.name() != 'tensorboard':
        continue
      cmd_args = p.cmdline()
      del cmd_args[0:2] # remove 'python' and 'tensorboard'
      args = parser.parse_args(cmd_args)
      running_list.append({'pid': p.pid, 'logdir': args.logdir, 'port': args.port})
    return running_list

  @staticmethod
  def get_reverse_proxy_url(port):
    """Get the reverse proxy url. Note that this URL only works with
       Datalab web server which supports reverse proxy.

    Args:
      port: the port of the tensorboard instance.
    Returns: the reverse proxy URL.
    """
    return '/_proxy/%d/' % port

  @staticmethod
  def start(logdir):
    """Start a TensorBoard instance.

    Args:
      logdir: the logdir to run TensorBoard on.
    Returns:
      A tuple. First is the pid of the instance. Second is the port used.
    Raises:
      Exception if the instance cannot be started.
    """
    if logdir.startswith('gs://'):
      # Check user does have access. TensorBoard will start successfully regardless
      # the user has read permissions or not so we check permissions here to
      # give user alerts if needed.
      datalab.storage._api.Api.verify_permitted_to_read(logdir)

    port = datalab.utils.pick_unused_port()
    args = ['tensorboard', '--logdir=' + logdir, '--port=' + str(port)]
    p = subprocess.Popen(args)
    retry = 5
    while (retry > 0):
      if datalab.utils.is_http_running_on(port):
        return p.pid, port
      time.sleep(1)
      retry -= 1

    raise Exception('Cannot start TensorBoard.')

  @staticmethod
  def stop(pid):
    """Shut down a specific process.

    Args:
      pid: the pid of the process to shutdown.
    """
    if psutil.pid_exists(pid):
      try:
        p = psutil.Process(pid)
        p.kill()
        p.wait()
      except Exception:
        pass
